#!/bin/bash

set -e

function usage {
	echo "Usage: $0 [--iso ISO] [--hashes ISO-Hashfile] [--filenames Filenames-File] -DEVICE"
	echo "Example: $0 --iso tails-i386-1.5.1.iso /dev/sdc"
	exit 1
}

function create_loop_devices {
        declare -g -a LOOP_MINOR_VERSIONS
        counter=3
        for i in {8..63}
        do
                if [ -e /dev/loop$i ]
                then
                        continue
                fi

                LOOP_MINOR_VERSIONS+=($i)
                mknod /dev/loop$i b 7 $i
                chown --reference=/dev/loop0 /dev/loop$i
                chmod --reference=/dev/loop0 /dev/loop$i

                counter=$((counter-1))
                if [ $counter -eq 0 ]
                        then break
                fi
        done
}

function remove_loop_devices {
        for i in "${LOOP_MINOR_VERSIONS[@]}"
        do
                rm /dev/loop$i
        done
}

function cleanup {
	set +e
	remove_loop_devicesSH_FILE="${2}"
            USE_EXISTING_HASH_FILE=true
            BASE_NAME=`basename "${HASH_FILE}"`
            shift # past argument
	umount ${MOUNT_POINT}/{squashfs,boot}
	rmdir ${MOUNT_POINT}/{squashfs,boot} ${MOUNT_POINT}
	END=$(date +%s.%N)
	DIFF=$(echo "$END - $START" | bc)
	echo "Runtime: $DIFF seconds"
}

[[ $# -lt 3 ]] && usage

while [[ $# > 0 ]]
do
	key="${1}"
	case ${key} in
	    --iso)
	    ISO="${2}"
	    BASE_NAME=`basename "${ISO}"`
	    shift # past argument
	    ;;
	    --hashes)
	    HASH_FILE="${2}"
	    USE_EXISTING_HASH_FILE=true
	    BASE_NAME=`basename "${HASH_FILE}"`
	    shift # past argument
	    ;;
	    --filenames)
            FILENAMES_FILE="${2}"
            USE_EXISTING_FILENAMES_FILE=true
            shift # past argument
	    ;;
	    *)
	    DEV="${1}"
	    break
	    ;;
	esac
	shift # past argument or value
done

echo "DEVICE: $DEV"
echo "ISO: $ISO"
echo "HASH_FILE: $HASH_FILE"
echo "FILENAMES_FILE: $FILENAMES_FILE"

[[ ! "${DEV}" || ( ! "${ISO}" && ! "${HASH_FILE}" ) ]] && usage

[[ ! `which hashrat` ]] && echo "Couldn't find hashrat" && exit 1
HASHRAT=`which hashrat`

START=$(date +%s.%N)
DIR="$( cd "$( dirname "${BASH_SOURCE[0]}" )" && pwd )"	
MOUNT_POINT="/tmp/tails-verifier"
DEFAULT_HASH_FILE="${DIR}/output/${BASE_NAME}_sha256.txt"
DEFAULT_FILENAMES_FILE="${DIR}/output/${BASE_NAME}_filenames.txt"
LOG_FILE="${DIR}/verify-${BASE_NAME}.log"
EXCLUDED="squashfs/var/lib/urandom/random-seed"


trap cleanup EXIT

echo "Creating mount points..."
create_loop_devices
mkdir ${MOUNT_POINT} ${MOUNT_POINT}/{boot,squashfs}

if [ ! "${USE_EXISTING_HASH_FILE}" ]
then
	HASH_FILE="${DEFAULT_HASH_FILE}"

	echo "Mounting ISO..."
	mount -o ro,loop ${ISO} ${MOUNT_POINT}/boot
	mount -o ro,loop ${MOUNT_POINT}/boot/live/filesystem.squashfs ${MOUNT_POINT}/squashfs/

	echo "Hashing ISO (storing hashes in ${HASH_FILE})..."
	set +e  # TODO: remove this when hashrat is fixed to return 0 on success
		# see https://github.com/ColumPaget/Hashrat/issues/1
	${HASHRAT} -r -sha256 ${MOUNT_POINT} 2>&1 > ${HASH_FILE} | grep -v 'Not following symbolic link'
	set -e

	echo "Renaming occurrences of boot/isolinux in hashfile to boot/syslinux"
	sed -i 's/boot\/isolinux/boot\/syslinux/g' "${HASH_FILE}"
	sed -i 's/syslinux\/isolinux.cfg/syslinux\/syslinux.cfg/g' "${HASH_FILE}"
	grep "EFI\/BOOT\/isolinux.cfg" "${HASH_FILE}" | sed 's/isolinux.cfg/syslinux.cfg/g' >> "${HASH_FILE}"	

	echo "Unmounting ISO..."
	umount ${MOUNT_POINT}/{squashfs,boot}
fi


if [ ! "${USE_EXISTING_FILENAMES_FILE}" ]
then
	FILENAMES_FILE="${DEFAULT_FILENAMES_FILE}"

	echo "Mounting ISO..."
	mount -o ro,loop ${ISO} ${MOUNT_POINT}/boot
	mount -o ro,loop ${MOUNT_POINT}/boot/live/filesystem.squashfs ${MOUNT_POINT}/squashfs/

	echo "Storing filenames in ${FILENAMES_FILE}"
	find ${MOUNT_POINT} -printf "%P\n" > "${FILENAMES_FILE}"
	
	echo "Renaming files in ${FILENAMES_FILE}"
	sed -i 's/boot\/isolinux/boot\/syslinux/g' "${FILENAMES_FILE}"
	sed -i 's/syslinux\/isolinux.cfg/syslinux\/syslinux.cfg/g' "${FILENAMES_FILE}"
	grep "EFI\/BOOT\/isolinux.cfg" "${FILENAMES_FILE}" | sed 's/isolinux.cfg/syslinux.cfg/g' >> "${FILENAMES_FILE}"
	sort "${FILENAMES_FILE}" -o "${FILENAMES_FILE}"

	echo "Unmounting ISO..."
	umount ${MOUNT_POINT}/{squashfs,boot}
fi


echo "Mounting Tails device..."
mount -o ro,loop "${DEV}1" "${MOUNT_POINT}/boot"
mount -o ro,loop "${MOUNT_POINT}/boot/live/filesystem.squashfs" "${MOUNT_POINT}/squashfs/"

set +e

echo "Verifying Tails device files (logging to ${LOG_FILE})..."
cat "${HASH_FILE}" | "${HASHRAT}" -c "${MOUNT_POINT}" -x "${EXCLUDED}" &> "${LOG_FILE}"

if [ $? != 0 ] 
then
	NUM_ERRORS=`grep -c -e "ERROR\|FAILED" "${LOG_FILE}"`
	echo "Verification failed with ${NUM_ERRORS} errors. See ${LOG_FILE} for a list of the mismatched files"
	exit 1
fi


echo "Checking for additional files on device (logging to ${LOG_FILE})"
find "${MOUNT_POINT}" -printf "%P\n" | sort | grep -v "^boot/syslinux/ldlinux.sys" | diff - "${FILENAMES_FILE}" >> "${LOG_FILE}"

if [ $? != 0 ]
then
	echo "Verification failed: There are files on the device which are not in the iso. See ${LOG_FILE} for further information"
	exit 1
fi


echo "Checking the Master Boot Record"
MBR="/tmp/mbr.bin"
ISO_MBR="${MOUNT_POINT}/squashfs/usr/lib/SYSLINUX/mbr.bin"
ISO_ALTMBR="${MOUNT_POINT}/squashfs/usr/lib/SYSLINUX/altmbr.bin"
ISO_GPTMBR="${MOUNT_POINT}/squashfs/usr/lib/SYSLINUX/gptmbr.bin"
HASH_ISO_MBR=`sha256sum < "${ISO_MBR}"`
HASH_ISO_ALTMBR=`sha256sum < "${ISO_ALTMBR}"`
HASH_ISO_GPTMBR=`sha256sum < "${ISO_GPTMBR}"`
echo "sha256sum ${ISO_MBR} ${HASH_ISO_MBR}" > "${LOG_FILE}"
echo "sha256sum ${ISO_ALTMBR} ${HASH_ISO_ALTMBR}" > "${LOG_FILE}"
echo "sha256sum ${ISO_GPTMBR} ${HASH_ISO_GPTMBR}" > "${LOG_FILE}"
MBRSIZE=`wc -c < "${ISO_MBR}"`
dd if="${DEV}" of="${MBR}" bs=1 count="${MBRSIZE}" 2> /dev/null
HASH_MBR=`sha256sum < "${MBR}"`
echo "sha256sum ${MBR} ${HASH_MBR}" > "${LOG_FILE}"
if 	[ "${HASH_MBR}" != "${HASH_ISO_MBR}" ] \
	&& [ "${HASH_MBR}" != "${HASH_ISO_ALTMBR}" ] \
	&& [ "${HASH_MBR}" != "${HASH_ISO_GPTMBR}" ]
then
	echo "Verification failed: The Master Boot Record doesn't match any of the MBRs on the ISO"
	exit 1
fi

echo "Verification successful"
exit 0
